package com.genpt.nsight.v4.service.impl;

import com.genpt.nsight.NSightCache;
import com.genpt.nsight.config.NSightProperties;
import com.genpt.nsight.v1.model.HeaderResponse;
import com.genpt.nsight.v1.model.QuantityFreePaid;
import com.genpt.nsight.v1.model.SiteRelationship;
import com.genpt.nsight.v1.model.dto.SiteDTO;
import com.genpt.nsight.v4.AvailabilityServiceContextV4;
import com.genpt.nsight.v4.SplunkLogger;
import com.genpt.nsight.v4.model.*;
import com.genpt.nsight.v4.service.ProductAvailabilityAggregateNOLV4;
import com.genpt.nsight.v4.service.ProductAvailabilityFreePaidServiceV4;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Service("ProductAvailabilityAggregateNOLV4")
public class ProductAvailabilityAggregateNOLV4Impl implements ProductAvailabilityAggregateNOLV4 {
    @Autowired
    NSightCache nSightCache;
    private static final Logger LOGGER = LoggerFactory.getLogger(ProductAvailabilityAggregateNOLV4Impl.class);

    @Autowired
    NSightProperties nSightProperties;

    private ProductAvailabilityFreePaidServiceV4 productAvailabilityFreePaidServiceV4;

    private String deliverOffset = "DeliverFromStore";
    private String onDemandOffset = "DeliverOnDemand";
    private String pickupOffset = "PickupFromStore";

    @Autowired
    public void setProductAvailabilityFreePaidServiceV4(ProductAvailabilityFreePaidServiceV4 productAvailabilityFreePaidServiceV4) {
        this.productAvailabilityFreePaidServiceV4 = productAvailabilityFreePaidServiceV4;
    }

    @Override
    public DetailResponseAggNOL setElementsFromRequest(AvailabilityServiceContextV4 availabilityServiceContextV4, ProductAvailabilityAggResponseNOL productAvailabilityAggResponseNOL){
        productAvailabilityAggResponseNOL.setCorrelationId(availabilityServiceContextV4.getCorrelationID());
        HeaderResponse headerResponse = new HeaderResponse();
        headerResponse.setCountryCode(availabilityServiceContextV4.getRequest().getHeaderRequest().getCountryCode());
        headerResponse.setLanguageCode(availabilityServiceContextV4.getRequest().getHeaderRequest().getLanguageCode());
        headerResponse.setRequestor(availabilityServiceContextV4.getRequest().getHeaderRequest().getRequestor());
        headerResponse.setResponder(availabilityServiceContextV4.getRequest().getHeaderRequest().getResponder());
        headerResponse.setTimestamp(Long.valueOf(availabilityServiceContextV4.getRequest().getHeaderRequest().getTimestamp()));
        productAvailabilityAggResponseNOL.setHeaderResponse(headerResponse);
        DetailResponseAggNOL detailResponseAggNOL= new DetailResponseAggNOL();
        if (availabilityServiceContextV4.getRequest().getDetailRequest().getDeliveryOffSets()!=null)
            detailResponseAggNOL.setDeliveryOffSets(availabilityServiceContextV4.getRequest().getDetailRequest().getDeliveryOffSets());
        if (availabilityServiceContextV4.getRequest().getDetailRequest().getSubBuckets()!=null)
            detailResponseAggNOL.setSubBuckets(availabilityServiceContextV4.getRequest().getDetailRequest().getSubBuckets());

        if (availabilityServiceContextV4.getZipCodesSet()!=null && availabilityServiceContextV4.getZipCodesSet().size() >0) {
            ShippingDestination shippingDestination = new ShippingDestination();
            shippingDestination.setZipcode(availabilityServiceContextV4.getZipCodesSet().iterator().next());
            detailResponseAggNOL.setShippingDestination(shippingDestination);
        }
         return  detailResponseAggNOL;
    }

    @Override
    public ProductAvailabilityAggResponseNOL getProductAvailability(AvailabilityServiceContextV4 availabilityServiceContextV4) {
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV4.getRequest().getHeaderRequest());
        long startTime = System.currentTimeMillis();

        String siteId = nSightCache.getAliasToSiteIdMap().get(availabilityServiceContextV4.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getType() + "|" + availabilityServiceContextV4.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getValue());
        SiteDTO callingStoreDTO = nSightCache.getSiteIdToSiteMap().get(siteId);

        if(callingStoreDTO.getTimeZone() == null || callingStoreDTO.getTimeZone().equalsIgnoreCase("NAN")) {
            com.genpt.nsight.v4.model.ProductAvailabilityAggResponseNOL productAvailabilityResponseFreePaid = availabilityServiceContextV4.getProductAvailabilityResponseNOL(availabilityServiceContextV4);
            LOGGER.error(splunkLogger.toString() + " Invalid Data - Timezone is null in NSight for requesting site: 3005");
            productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV4.createInvalidTimezoneTracking());
            return productAvailabilityResponseFreePaid;
        }
        if (nSightCache.getStoreHoursList().get(callingStoreDTO.getSiteIdentifier()) == null){
            com.genpt.nsight.v4.model.ProductAvailabilityAggResponseNOL productAvailabilityResponseFreePaid = availabilityServiceContextV4.getProductAvailabilityResponseNOL(availabilityServiceContextV4);
            LOGGER.error(splunkLogger.toString() + " Invalid Data - No Store hours is null in NSight for requesting site: 3005");
            productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV4.createInvalidStoreHoursTracking());
            return productAvailabilityResponseFreePaid;
        }
        if(callingStoreDTO.getTimeZoneID() == null ) {
            com.genpt.nsight.v4.model.ProductAvailabilityAggResponseNOL productAvailabilityResponseFreePaid = availabilityServiceContextV4.getProductAvailabilityResponseNOL(availabilityServiceContextV4);
            LOGGER.error(splunkLogger.toString() + " Invalid Data - Timezone ID is null in NSight for requesting site: 3005");
            productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV4.createInvalidTimezoneIDTracking());
            return productAvailabilityResponseFreePaid;
        }

        if(availabilityServiceContextV4.getTimeOfDay() == 0L) {
            availabilityServiceContextV4.setDayOfWeekLocalStore(callingStoreDTO);
        }

        long buildResponseStartTime = System.currentTimeMillis();
        ProductAvailabilityAggResponseNOL productAvailabilityAggResponseNOL = new ProductAvailabilityAggResponseNOL();
        DetailResponseAggNOL detailResponseAggNOL = setElementsFromRequest(availabilityServiceContextV4, productAvailabilityAggResponseNOL);

        productAvailabilityAggResponseNOL.setDetailResponse(detailResponseAggNOL);
        //call the V4 free paid service
        ProductAvailabilityResponseFreePaid productAvailabilityResponseFreePaid = productAvailabilityFreePaidServiceV4.getProductAvailability(availabilityServiceContextV4, false);
        //based on the free paid derive the aggregates
        //go over all the parts
        productAvailabilityAggResponseNOL.setTracking(productAvailabilityResponseFreePaid.getTracking());

        productAvailabilityResponseFreePaid.getDetailResponse().getParts().forEach(part -> {// at each part level
            if (part.getQuantities()!=null ){
                Map<String, Bucket> bucketMap = new HashMap<>();//map of buckename to Bucket(LocalStore, Today, etc)

                List<PartAggNOL> partAggNOLs =  productAvailabilityAggResponseNOL.getDetailResponse().getParts();
                if (partAggNOLs == null) {
                    partAggNOLs = new ArrayList<PartAggNOL>();
                    productAvailabilityAggResponseNOL.getDetailResponse().setParts(partAggNOLs);
                }
                //populate this part
                PartAggNOL partAggNOL = new PartAggNOL();
                partAggNOL.setLineAbbrev(part.getLineAbbrev());
                partAggNOL.setPartNumber(part.getPartNumber());
                partAggNOL.setTracking(part.getTracking());
                if (part.getPrice()!=null)
                    partAggNOL.setPrice(part.getPrice());
                if (part.getProductKit()!=null)
                    partAggNOL.setProductKit(part.getProductKit());
                if (part.getSuperSedes()!=null)
                    partAggNOL.setSuperSedes(part.getSuperSedes());

                BucketQuantity bucketQuantity= partAggNOL.getQuantities();
                if (bucketQuantity==null){//local bucket does not exist
                    bucketQuantity = new BucketQuantity();
                    bucketQuantity.setBuckets(new ArrayList<>());
                    partAggNOL.setQuantities(bucketQuantity);
                }

                // now iterate quantities from each site to create the buckets
                for (QuantityFreePaid quantityFreePaid : part.getQuantities()){
                    if (quantityFreePaid.getAvailable() != null && quantityFreePaid.getAvailable().intValue() > 0) {
                        //Create the Local Store bucket if calling store
                        String currentSiteID = quantityFreePaid.getSiteUUID();
                        if (currentSiteID.equals(availabilityServiceContextV4.getRequestingSite().getSiteIdentifier())) {//local store
                            createLocalBucket(bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                        }
                        else if (availabilityServiceContextV4.getRequestingSite().getMainCounterFlag()==true && availabilityServiceContextV4.getPrimaryDC() !=null && currentSiteID.compareToIgnoreCase(availabilityServiceContextV4.getPrimaryDC().getSiteId()) ==0) {
                            //store is a main counter so add the DC quantity to local bucket
                            createLocalBucket(bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                            //now if this DC is NOL NXP Supported Fulfillments then this quanity goes towards that bucket too
                            //so check if it has that relationship
                            List<SiteRelationship> siteRelationshipList = availabilityServiceContextV4.getSiteRelationMap().get(currentSiteID);
                            for (SiteRelationship siteRelationDTO : siteRelationshipList) {
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("NOL NXP Supported Fulfillments") == 0) {
                                    checkShipToHomeDeliveries( availabilityServiceContextV4,  currentSiteID,  bucketMap, bucketQuantity,  quantityFreePaid);
                                    break;
                                }
                            }
                        }
                        else {
                            //first check if it is also
                            //we will check for store console inventory relationships
                            List<SiteRelationship> siteRelationshipList = availabilityServiceContextV4.getSiteRelationMap().get(currentSiteID);

                            for (SiteRelationship siteRelationDTO : siteRelationshipList) {
                                //check if its a store console friend
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("Store Console Inventory Friend") == 0) {
                                    // add it to the appropriate bucket. lets get the deliveries for this site
                                    List<com.genpt.nsight.v4.model.Delivery> deliveryList = availabilityServiceContextV4.getDeliveryBySiteSorted().get(currentSiteID);
                                    if (deliveryList != null && deliveryList.size() > 0) {// has deliveries. ideally we should be interested in only the first one
                                        //figure out which bucket it goes into // add it to the appropriate bucket
                                        createBucketsFromDelivery(  availabilityServiceContextV4,deliveryList, bucketMap ,quantityFreePaid,bucketQuantity, null) ;
                                    }
                                }//end of store console friend

                                // we would need the data from site to zip for this
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("NOL NXP Supported Fulfillments") == 0) {
                                    checkShipToHomeDeliveries( availabilityServiceContextV4,  currentSiteID,  bucketMap, bucketQuantity,  quantityFreePaid);
                                }//end of NOL NXP Supported Fulfillments

                                //Nearby Stores bucket which doesnt use any delivery data
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("NOL Near By Stores") == 0) {
                                    //regardlessof deliver we create this bucket
                                    createNearByStores( bucketMap , bucketQuantity, quantityFreePaid);
                                }
                            }//end of relationship loop
                        }//else , not localstore
                    }else{
                        LOGGER.debug(splunkLogger.toString() + " Dropping Site {} since quantity is {} ", quantityFreePaid.getSiteUUID(), quantityFreePaid.getAvailable() );
                    }

                }
                partAggNOLs.add(partAggNOL);
            }//valid part
        });//for each part

        int totalPartsRequested = getTotalPartsRequested(availabilityServiceContextV4);
        LOGGER.debug(splunkLogger.toString() + " Time taken to build NOLAggregate Impl V4 response = {} and the total parts requested are = {} ",  (System.currentTimeMillis() - buildResponseStartTime), totalPartsRequested);
        LOGGER.info(splunkLogger.toString() + " Total time within NOLAggregate Impl V4 MicroService = {} and the total parts requested are = {} ",(System.currentTimeMillis() - startTime), totalPartsRequested);

        return productAvailabilityAggResponseNOL;
    }

    private static int getTotalPartsRequested(AvailabilityServiceContextV4 availabilityServiceContextV4) {
        if(availabilityServiceContextV4 != null && availabilityServiceContextV4.getRequest() != null
                && availabilityServiceContextV4.getRequest().getDetailRequest() != null && availabilityServiceContextV4.getRequest().getDetailRequest().getParts() != null) {
            return availabilityServiceContextV4.getRequest().getDetailRequest().getParts().size();
        }
        return 0;
    }

    @Override
    public ProductAvailabilityAggResponseNOL getProductAvailabilityV4V2(AvailabilityServiceContextV4 availabilityServiceContextV4) {
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV4.getRequest().getHeaderRequest());
        long startTime = System.currentTimeMillis();

        String siteId = nSightCache.getAliasToSiteIdMap().get(availabilityServiceContextV4.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getType() + "|" + availabilityServiceContextV4.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getValue());
        SiteDTO callingStoreDTO = nSightCache.getSiteIdToSiteMap().get(siteId);

        if(callingStoreDTO.getTimeZone() == null || callingStoreDTO.getTimeZone().equalsIgnoreCase("NAN")) {
            com.genpt.nsight.v4.model.ProductAvailabilityAggResponseNOL productAvailabilityResponseFreePaid = availabilityServiceContextV4.getProductAvailabilityResponseNOL(availabilityServiceContextV4);
            LOGGER.error(splunkLogger.toString() + " Invalid Data - Timezone is null in NSight for requesting site: 3005");
            productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV4.createInvalidTimezoneTracking());
            return productAvailabilityResponseFreePaid;
        }
        if (nSightCache.getStoreHoursList().get(callingStoreDTO.getSiteIdentifier()) == null){
            com.genpt.nsight.v4.model.ProductAvailabilityAggResponseNOL productAvailabilityResponseFreePaid = availabilityServiceContextV4.getProductAvailabilityResponseNOL(availabilityServiceContextV4);
            LOGGER.error(splunkLogger.toString() + " Invalid Data - No Store hours is null in NSight for requesting site: 3005");
            productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV4.createInvalidStoreHoursTracking());
            return productAvailabilityResponseFreePaid;
        }
        if(callingStoreDTO.getTimeZoneID() == null ) {
            com.genpt.nsight.v4.model.ProductAvailabilityAggResponseNOL productAvailabilityResponseFreePaid = availabilityServiceContextV4.getProductAvailabilityResponseNOL(availabilityServiceContextV4);
            LOGGER.error(splunkLogger.toString() + " Invalid Data - Timezone ID is null in NSight for requesting site: 3005");
            productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV4.createInvalidTimezoneIDTracking());
            return productAvailabilityResponseFreePaid;
        }

        if(availabilityServiceContextV4.getTimeOfDay() == 0L) {
            availabilityServiceContextV4.setDayOfWeekLocalStore(callingStoreDTO);
        }

        long buildResponseStartTime = System.currentTimeMillis();
        ProductAvailabilityAggResponseNOL productAvailabilityAggResponseNOL = new ProductAvailabilityAggResponseNOL();
        DetailResponseAggNOL detailResponseAggNOL = setElementsFromRequest(availabilityServiceContextV4, productAvailabilityAggResponseNOL);

        productAvailabilityAggResponseNOL.setDetailResponse(detailResponseAggNOL);
        //call the V4 free paid service
        ProductAvailabilityResponseFreePaid productAvailabilityResponseFreePaid = productAvailabilityFreePaidServiceV4.getProductAvailability(availabilityServiceContextV4, false);
        //based on the free paid derive the aggregates
        //go over all the parts
        productAvailabilityAggResponseNOL.setTracking(productAvailabilityResponseFreePaid.getTracking());

        productAvailabilityResponseFreePaid.getDetailResponse().getParts().forEach(part -> {// at each part level
            if (part.getQuantities()!=null ){
                Map<String, Bucket> bucketMap = new HashMap<>();//map of buckename to Bucket(LocalStore, Today, etc)

                List<PartAggNOL> partAggNOLs =  productAvailabilityAggResponseNOL.getDetailResponse().getParts();
                if (partAggNOLs == null) {
                    partAggNOLs = new ArrayList<PartAggNOL>();
                    productAvailabilityAggResponseNOL.getDetailResponse().setParts(partAggNOLs);
                }
                //populate this part
                PartAggNOL partAggNOL = new PartAggNOL();
                partAggNOL.setLineAbbrev(part.getLineAbbrev());
                partAggNOL.setPartNumber(part.getPartNumber());
                partAggNOL.setTracking(part.getTracking());
                if (part.getPrice()!=null)
                    partAggNOL.setPrice(part.getPrice());
                if (part.getProductKit()!=null)
                    partAggNOL.setProductKit(part.getProductKit());
                if (part.getSuperSedes()!=null)
                    partAggNOL.setSuperSedes(part.getSuperSedes());

                BucketQuantity bucketQuantity= partAggNOL.getQuantities();
                if (bucketQuantity==null){//local bucket does not exist
                    bucketQuantity = new BucketQuantity();
                    bucketQuantity.setBuckets(new ArrayList<>());
                    partAggNOL.setQuantities(bucketQuantity);
                }

                // now iterate quantities from each site to create the buckets
                for (QuantityFreePaid quantityFreePaid : part.getQuantities()){
                    if (quantityFreePaid.getAvailable() != null && quantityFreePaid.getAvailable().intValue() > 0) {
                        //Create the Local Store bucket if calling store
                        String currentSiteID = quantityFreePaid.getSiteUUID();
                        if (currentSiteID.equals(availabilityServiceContextV4.getRequestingSite().getSiteIdentifier())) {//local store
                            createLocalBucket(bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                        }
                        else if (availabilityServiceContextV4.getRequestingSite().getMainCounterFlag()==true && availabilityServiceContextV4.getPrimaryDC() !=null && currentSiteID.compareToIgnoreCase(availabilityServiceContextV4.getPrimaryDC().getSiteId()) ==0) {
                            //store is a main counter so add the DC quantity to local bucket
                            createLocalBucket(bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                            //now if this DC is NOL NXP Supported Fulfillments then this quanity goes towards that bucket too
                            //so check if it has that relationship
                            List<SiteRelationship> siteRelationshipList = availabilityServiceContextV4.getSiteRelationMap().get(currentSiteID);
                            for (SiteRelationship siteRelationDTO : siteRelationshipList) {
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("NOL NXP Supported Fulfillments") == 0) {
                                    checkShipToHomeDeliveries( availabilityServiceContextV4,  currentSiteID,  bucketMap, bucketQuantity,  quantityFreePaid);
                                    break;
                                }
                            }
                        }
                        else {
                            //we will check for store console inventory relationships
                            List<SiteRelationship> siteRelationshipList = availabilityServiceContextV4.getSiteRelationMap().get(currentSiteID);

                            for (SiteRelationship siteRelationDTO : siteRelationshipList) {
                                //check if its a store console friend
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("Store Console Inventory Friend") == 0) {
                                    // add it to the appropriate bucket. lets get the deliveries for this site
                                    List<com.genpt.nsight.v4.model.Delivery> deliveryList = availabilityServiceContextV4.getDeliveryBySiteSorted().get(currentSiteID);
                                    if (deliveryList != null && deliveryList.size() > 0) {// has deliveries. ideally we should be interested in only the first one
                                        //figure out which bucket it goes into // add it to the appropriate bucket
                                        if (availabilityServiceContextV4.getDeliverFromStoreOffsetRaw() ==null && availabilityServiceContextV4.getDeliverOnDemandOffsetRaw()==null && availabilityServiceContextV4.getPickUpFromStoreOffsetRaw() ==null)
                                            createBucketsFromDelivery(  availabilityServiceContextV4,deliveryList, bucketMap ,quantityFreePaid,bucketQuantity, null) ;
                                        else{
                                            if (availabilityServiceContextV4.getDeliverFromStoreOffsetRaw()!=null)
                                                createBucketsFromDelivery(  availabilityServiceContextV4,deliveryList, bucketMap ,quantityFreePaid,bucketQuantity, deliverOffset) ;
                                            if (availabilityServiceContextV4.getDeliverOnDemandOffsetRaw()!=null)
                                                createBucketsFromDelivery(  availabilityServiceContextV4,deliveryList, bucketMap ,quantityFreePaid,bucketQuantity, onDemandOffset) ;
                                            if (availabilityServiceContextV4.getPickUpFromStoreOffsetRaw()!=null)
                                                createBucketsFromDelivery(  availabilityServiceContextV4,deliveryList, bucketMap ,quantityFreePaid,bucketQuantity, pickupOffset) ;
                                        }
                                    }
                                }//end of store console friend

                                // we would need the data from site to zip for this
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("NOL NXP Supported Fulfillments") == 0) {
                                    checkShipToHomeDeliveries( availabilityServiceContextV4,  currentSiteID,  bucketMap, bucketQuantity,  quantityFreePaid);
                                }//end of NOL NXP Supported Fulfillments

                                //Nearby Stores bucket which doesnt use any delivery data
                                if (siteRelationDTO.getSiteRelationshipType().compareToIgnoreCase("NOL Near By Stores") == 0) {
                                    //regardlessof deliver we create this bucket
                                    createNearByStores( bucketMap , bucketQuantity, quantityFreePaid);
                                }
                            }//end of relationship loop
                        }//else , not localstore
                    }else{
                        LOGGER.debug(splunkLogger.toString() + " Dropping Site {} since quantity is {} ", quantityFreePaid.getSiteUUID(), quantityFreePaid.getAvailable() );
                    }

                }
                partAggNOLs.add(partAggNOL);
            }//valid part
        });//for each part

        int totalPartsRequested = getTotalPartsRequested(availabilityServiceContextV4);
        LOGGER.debug(splunkLogger.toString() + " Time taken to build NOLAggregate Impl V4.2 response = {} and the total parts requested are = {}",  (System.currentTimeMillis() - buildResponseStartTime), totalPartsRequested);
        LOGGER.info(splunkLogger.toString() + " Total time within NOLAggregate Impl V4.2 MicroService = {} and the total parts requested are = {}",(System.currentTimeMillis() - startTime), totalPartsRequested);

        return productAvailabilityAggResponseNOL;
    }

    private void createBucketsFromDelivery(AvailabilityServiceContextV4 availabilityServiceContextV4, List<com.genpt.nsight.v4.model.Delivery> deliveryList,
                                           Map<String, Bucket> bucketMap , QuantityFreePaid quantityFreePaid, BucketQuantity bucketQuantity, String offsetName) {
        //once its in a bucket we dont need to look at other deliveries
        for (com.genpt.nsight.v4.model.Delivery delivery : deliveryList) {//iterate deliveries
            //check for today1 and today2 buckets
            // this is only for free options
            if (delivery.getCutOffDateTime() != null && delivery.isFreeDelivery()) {//valid free delivery
                ZonedDateTime quickest = null;

                if (offsetName == null)
                    quickest   = delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin();
                else if (offsetName.compareToIgnoreCase(deliverOffset)==0)
                    quickest = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMinDeliver(), false);
                else if (offsetName.compareToIgnoreCase(pickupOffset)==0)
                    quickest = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMinPickup(), false);
                else if (offsetName.compareToIgnoreCase(onDemandOffset)==0)
                    quickest= availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMinOnDemand(), false);

                //the call is made within store hours. need to consider after applying the offsets
                if (!availabilityServiceContextV4.isAfterStoreHours()){
                    // do we have any deliveries that can be done today. also check if its within after applying the offset
                    if (availabilityServiceContextV4.getToday1CutoffTime() != null) {//can have two subbucekts
                        if (quickest.isBefore(availabilityServiceContextV4.getToday1CutoffTime().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset()))) {
                            //can be got in the first Today1 subbucket for day
                            if (offsetName==null)
                                createOrUpdateTodayTomorrowBucket( delivery, "Today1", bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                            else
                                createOrUpdateTodayTomorrowBucket( delivery, "Today1"+offsetName, bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                            break;//from this delivery
                        } else {//after today1 cutofftime
                            if (quickest.isBefore(availabilityServiceContextV4.getStoreClose().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset()))
                            || quickest.isEqual(availabilityServiceContextV4.getStoreClose().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset()))) {
                                //goes into Today2 sub bucket
                                if (offsetName==null)
                                    createOrUpdateTodayTomorrowBucket(delivery, "Today2", bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                                else
                                    createOrUpdateTodayTomorrowBucket(delivery, "Today2"+offsetName, bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                                break;
                            } else {//its after store close. so could be tomorrow ot plus2 bucket
                                //check for tomorrow cutoff offset
                                insertIntoTomorrowOrPlus2(delivery, availabilityServiceContextV4, bucketMap, quantityFreePaid, bucketQuantity,offsetName);
                                break;
                            }
                        }
                    } else {//can have only one sub bucket today
                        if ((quickest.isBefore(availabilityServiceContextV4.getStoreClose().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset()))
                                || quickest.isEqual(availabilityServiceContextV4.getStoreClose().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())))
                                && !availabilityServiceContextV4.isStoreHolidayToday()
                        ) {
                            if (offsetName==null)
                                createOrUpdateTodayTomorrowBucket(delivery, "Today", bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                            else
                                createOrUpdateTodayTomorrowBucket(delivery, "Today"+offsetName, bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                            break;
                        } else {//can be tomorrow or plus2 bucket
                            insertIntoTomorrowOrPlus2(delivery, availabilityServiceContextV4, bucketMap, quantityFreePaid, bucketQuantity,offsetName);
                            break;
                        }
                    }
                } else {//store closed, can have only tomorrow abd plus 2
                    insertIntoTomorrowOrPlus2(delivery, availabilityServiceContextV4, bucketMap, quantityFreePaid, bucketQuantity,offsetName);
                    break;
                }
            }//valid cutofftime
        }//each delivery
    }

    private void createNearByStores( Map<String, Bucket> bucketMap,BucketQuantity bucketQuantity, QuantityFreePaid quantityFreePaid){
        //regardlessof deliver we create this bucket
        String buckeName = "Nearby Stores";
        Bucket bucket = bucketMap.get(buckeName);
        if (bucket == null) {
            bucket = new Bucket();
            bucket.setType(buckeName);
            List<BucketFields> bucketFieldses = new ArrayList<BucketFields>();
            BucketFields bucketFields = new BucketFields();
            bucketFields.setName("availableQty");
            bucketFields.setValue(String.valueOf(quantityFreePaid.getAvailable().intValue()));
            bucketFieldses.add(bucketFields);
            bucket.setFields(bucketFieldses);
            bucketMap.put(buckeName, bucket);
            bucketQuantity.getBuckets().add(bucket);
        } else {
            List<BucketFields> bucketFieldses = bucket.getFields();
            for (BucketFields bucketFields : bucketFieldses) {
                if (bucketFields.getName().compareToIgnoreCase("availableQty") == 0) {
                    BigDecimal bigDecimal = quantityFreePaid.getAvailable().add(new BigDecimal(bucketFields.getValue()));
                    bucketFields.setValue(String.valueOf(bigDecimal.intValue()));
                }
            }
        }

    }

    private void checkShipToHomeDeliveries(AvailabilityServiceContextV4 availabilityServiceContextV4, String currentSiteID, Map<String, Bucket> bucketMap,BucketQuantity bucketQuantity, QuantityFreePaid quantityFreePaid){
        List<com.genpt.nsight.v4.model.Delivery> deliveryList = availabilityServiceContextV4.getDeliveryByZipSorted().get(currentSiteID);
        if (deliveryList != null && deliveryList.size() > 0) {// has deliveries. we should be interested in only the first one
            for (com.genpt.nsight.v4.model.Delivery delivery : deliveryList) {//iterate deliveries
                //only for Paid options
                if (delivery.getCutOffDateTime() != null && nSightProperties.getPaidMOTS().contains(delivery.getDeliveryType())) {
                    ZonedDateTime expectedMax =availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMax(), false);
                    ZonedDateTime expectedMin = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMin(), false);
                    ZonedDateTime expectedCO = availabilityServiceContextV4.parseDateTime(delivery.getCutOffDateTime(), false);
                    // all these contribute to the ship to home bucket.
                    createShipToHomeBucket(bucketMap, bucketQuantity,expectedCO, expectedMax, expectedMin, quantityFreePaid, availabilityServiceContextV4);
                    break; //one site can contribute only once to a bucket
                }//have a valid delivery
            }//each delivery
        }//valid deliveries exist
    }

    private void createShipToHomeBucket(Map<String, Bucket> bucketMap,BucketQuantity bucketQuantity ,ZonedDateTime expectedCO,ZonedDateTime expectedMax, ZonedDateTime expectedMin,QuantityFreePaid quantityFreePaid , AvailabilityServiceContextV4 availabilityServiceContextV4){
        String buckeName = "Ship to Home";
        Bucket bucket = bucketMap.get(buckeName);
        if (bucket==null) {
            bucket = new Bucket();
            bucket.setType(buckeName);
            bucket.setBucketCutoffTime(expectedCO);
            bucket.setExpectedMax(expectedMax);
            bucket.setExpectedMin(expectedMin);
            //add the fields
            List<BucketFields> bucketFieldses = new ArrayList<BucketFields>();
            BucketFields bucketFields = new BucketFields();
            bucketFields.setName("availableQty");
            bucketFields.setValue(String.valueOf(quantityFreePaid.getAvailable().intValue()));
            bucketFieldses.add(bucketFields);

            BucketFields bucketFieldsCutoff = new BucketFields();
            bucketFieldsCutoff.setName("cuttoffDateTime");
            bucketFieldsCutoff.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(expectedCO, 0));
            bucketFieldses.add(bucketFieldsCutoff);

            BucketFields bucketFieldsdeliverDateTimeMin = new BucketFields();
            bucketFieldsdeliverDateTimeMin.setName("deliverDateTimeMin");
            bucketFieldsdeliverDateTimeMin.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(expectedMin, 0));
            bucketFieldses.add(bucketFieldsdeliverDateTimeMin);

            BucketFields bucketFieldsdeliverDateTimeMax = new BucketFields();
            bucketFieldsdeliverDateTimeMax.setName("deliverDateTimeMax");
            bucketFieldsdeliverDateTimeMax.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(expectedMax, 0));
            bucketFieldses.add(bucketFieldsdeliverDateTimeMax);

            //set the list to the bucket
            bucket.setFields(bucketFieldses);
            bucketMap.put(buckeName, bucket);
            bucketQuantity.getBuckets().add(bucket);

        }else{// nneed to update a bucket
            List<BucketFields> bucketFieldses = bucket.getFields();
            for ( BucketFields bucketFields : bucketFieldses){
                if (bucketFields.getName().compareToIgnoreCase("availableQty")== 0){
                    BigDecimal bigDecimal =  quantityFreePaid.getAvailable().add(new BigDecimal(bucketFields.getValue()));
                    bucketFields.setValue(String.valueOf(bigDecimal.intValue()));
                }

                //also need to update the times if we find a better range
                if (bucketFields.getName().compareToIgnoreCase("cuttoffDateTime")== 0){
                    //change it if needed to
                    if (expectedCO.isBefore(bucket.getBucketCutoffTime())){
                        bucket.setBucketCutoffTime(expectedCO);
                        bucketFields.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(expectedCO, 0));
                    }
                }
                //we have to update the min and max too if these are later than the existng one
                if (bucketFields.getName().compareToIgnoreCase("deliverDateTimeMin")== 0){
                    //change it if needed to
                    if (expectedMin.isBefore(bucket.getExpectedMin())){
                        bucketFields.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(expectedMin, 0));
                        bucket.setExpectedMin(expectedMin);
                    }
                }
                if (bucketFields.getName().compareToIgnoreCase("deliverDateTimeMax")== 0){
                    //change it if needed to
                    if (expectedMax.isAfter(bucket.getExpectedMax())){
                        bucketFields.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(expectedMax, 0));
                        bucket.setExpectedMax(expectedMax);
                    }
                }
            }//iterated all fields for the bucket
        }//updating a bucket


    }

    private void insertIntoTomorrowOrPlus2(Delivery delivery, AvailabilityServiceContextV4 availabilityServiceContextV4, Map<String, Bucket> bucketMap, QuantityFreePaid quantityFreePaid, BucketQuantity bucketQuantity, String offsetName){
        //its after store close. so could be tomorrow ot plus2 bucket
        //check for tomorrow cutoff offset
        if (availabilityServiceContextV4.getTomorrowTime()!=null){//can have two tomorrow subbuckets
            //can be in tomorrows bucket only if store is open tomorrow
            if (availabilityServiceContextV4.getTomorrow1CutoffTime() !=null &&
                    (delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().isBefore(availabilityServiceContextV4.getTomorrow1CutoffTime().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())) ||
                            delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().equals(availabilityServiceContextV4.getTomorrow1CutoffTime().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())))
            ){
                if (offsetName==null)
                    createOrUpdateTodayTomorrowBucket(delivery, "Tomorrow1",bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                else
                    createOrUpdateTodayTomorrowBucket(delivery, "Tomorrow1"+offsetName,bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                return;
            }else{
                //check if store is open tomorrow to be able to add to tomorrow2
                if (availabilityServiceContextV4.getStoreCloseTomorrow()!=null &&
                        (delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().isBefore(availabilityServiceContextV4.getStoreCloseTomorrow().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())) ||
                                delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().equals(availabilityServiceContextV4.getStoreCloseTomorrow().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())))
                ){
                    if (offsetName==null)
                        createOrUpdateTodayTomorrowBucket(delivery, "Tomorrow2",bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                    else
                        createOrUpdateTodayTomorrowBucket(delivery, "Tomorrow2"+offsetName,bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                    return;
                }else{//goes into plus2
                    if(offsetName==null)
                        createOrUpdateTodayTomorrowBucket(delivery, "Plus2days",bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                    else
                        createOrUpdateTodayTomorrowBucket(delivery, "Plus2days"+offsetName,bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                    return;
                }
            }
        }else{//can have only tomorrow bucket
            if (availabilityServiceContextV4.getStoreCloseTomorrow()!=null &&
                    (delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().isAfter(availabilityServiceContextV4.getStoreOpenTomorrow().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset()))
                            || delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().equals(availabilityServiceContextV4.getStoreOpenTomorrow().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())))
                    && (delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().isBefore(availabilityServiceContextV4.getStoreCloseTomorrow().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())) ||
                    delivery.getDeliveryEstimate().getQuickestExpectationOfAllMin().equals(availabilityServiceContextV4.getStoreCloseTomorrow().withZoneSameLocal(availabilityServiceContextV4.getStoreZoneOffset())))
            ){
                if (offsetName==null)
                    createOrUpdateTodayTomorrowBucket(delivery, "Tomorrow",bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                else
                    createOrUpdateTodayTomorrowBucket(delivery, "Tomorrow"+offsetName,bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);

                return;
            }else{//goes into plus2days
                if (offsetName==null)
                    createOrUpdateTodayTomorrowBucket(delivery, "Plus2days",bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);
                else
                    createOrUpdateTodayTomorrowBucket(delivery, "Plus2days"+offsetName,bucketMap, quantityFreePaid, availabilityServiceContextV4, bucketQuantity);

                return;
            }
        }
    }

    private void createOrUpdateTodayTomorrowBucket(Delivery delivery, String bucketName, Map<String, Bucket> bucketMap, QuantityFreePaid quantityFreePaid, AvailabilityServiceContextV4 availabilityServiceContextV4 ,BucketQuantity bucketQuantity){
        //this is a local store. if bucket does not exist create it
        Bucket bucket = bucketMap.get(bucketName);

        ZonedDateTime expectedMax =availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMax(), false);
        ZonedDateTime expectedMin = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMin(), false);
        ZonedDateTime bucketCutoffTime = availabilityServiceContextV4.parseDateTime(delivery.getCutOffDateTime(), false);

        ZonedDateTime expectedMaxDeliver = null;
        if (delivery.getDeliveryEstimate().getExpectationRangeMaxDeliver()!=null) {
            expectedMaxDeliver = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMaxDeliver(), false);
        }
        ZonedDateTime expectedMinDeliver = null;
        if (delivery.getDeliveryEstimate().getExpectationRangeMinDeliver()!=null) {
            expectedMinDeliver = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMinDeliver(), false);
        }
        ZonedDateTime expectedMaxPickup = null;
        if (delivery.getDeliveryEstimate().getExpectationRangeMaxPickup()!=null) {
            expectedMaxPickup = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMaxPickup(), false);
        }
        ZonedDateTime expectedMinPickup = null;
        if (delivery.getDeliveryEstimate().getExpectationRangeMinPickup()!= null) {
            expectedMinPickup = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMinPickup(), false);
        }
        ZonedDateTime expectedMaxOndemand = null;
        if (delivery.getDeliveryEstimate().getExpectationRangeMaxOnDemand()!=null) {
            expectedMaxOndemand = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMaxOnDemand(), false);
        }
        ZonedDateTime expectedMinOnDemand = null;
        if (delivery.getDeliveryEstimate().getExpectationRangeMinOnDemand()!= null) {
            expectedMinOnDemand = availabilityServiceContextV4.parseDateTime(delivery.getDeliveryEstimate().getExpectationRangeMinOnDemand(), false);
        }

        if (bucket==null){
            bucket= new Bucket();
            bucket.setType(bucketName);
            bucket.setBucketCutoffTime(bucketCutoffTime);
            bucket.setExpectedMax(expectedMax);
            bucket.setExpectedMin(expectedMin);
            if (expectedMaxDeliver!=null)
                bucket.setExpectedMaxDeliver(expectedMaxDeliver);
            if (expectedMinDeliver!=null)
                bucket.setExpectedMinDeliver(expectedMinDeliver);
            if (expectedMaxPickup!=null)
                bucket.setExpectedMaxPickup(expectedMaxPickup);
            if (expectedMinPickup!=null)
                bucket.setExpectedMinPickup(expectedMinPickup);
            if (expectedMaxOndemand!=null)
                bucket.setExpectedMaxonDemand(expectedMaxOndemand);
            if (expectedMinOnDemand!=null)
                bucket.setExpectedMinOnDemand(expectedMinOnDemand);

            //first create the list for the fields
            List<BucketFields> bucketFieldses = new ArrayList<BucketFields>();

            //now create the fields for the bucket
            BucketFields bucketFields = new BucketFields();
            bucketFields.setName("availableQty");
            bucketFields.setValue(String.valueOf(quantityFreePaid.getAvailable().intValue()));
            bucketFieldses.add(bucketFields);

            BucketFields bucketFieldsCutoff = new BucketFields();
            bucketFieldsCutoff.setName("cuttoffDateTime");
            bucketFieldsCutoff.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(bucketCutoffTime, 0));
            bucketFieldses.add(bucketFieldsCutoff);

            if (bucketName.contains("Plus2days") || bucketName.contains("Today")|| bucketName.contains("Tomorrow")){// need the additional Mins
                if (expectedMinDeliver !=null) {
                    BucketFields bucketFieldsdeliverDateTimeMin = new BucketFields();
                    bucketFieldsdeliverDateTimeMin.setName("deliverDateTimeMin");
                    bucketFieldsdeliverDateTimeMin.setValue(delivery.getDeliveryEstimate().getExpectationRangeMinDeliver());
                    bucketFieldses.add(bucketFieldsdeliverDateTimeMin);
                }
                if (expectedMinPickup!=null){
                    BucketFields bucketFieldspickupDateTimeMin = new BucketFields();
                    bucketFieldspickupDateTimeMin.setName("pickupDateTimeMin");
                    bucketFieldspickupDateTimeMin.setValue(delivery.getDeliveryEstimate().getExpectationRangeMinPickup());
                    bucketFieldses.add(bucketFieldspickupDateTimeMin);
                }
            }

            if (expectedMaxPickup!=null) {
                BucketFields bucketFieldsPickup = new BucketFields();
                bucketFieldsPickup.setName("pickupDateTimeMax");
                bucketFieldsPickup.setValue(delivery.getDeliveryEstimate().getExpectationRangeMaxPickup());
                bucketFieldses.add(bucketFieldsPickup);
            }
            if (expectedMaxDeliver !=null) {
                BucketFields bucketFieldsDeliver = new BucketFields();
                bucketFieldsDeliver.setName("deliverDateTimeMax");
                bucketFieldsDeliver.setValue(delivery.getDeliveryEstimate().getExpectationRangeMaxDeliver());
                bucketFieldses.add(bucketFieldsDeliver);
            }
            //set the list to the bucket
            bucket.setFields(bucketFieldses);
            bucketMap.put(bucketName, bucket);
            bucketQuantity.getBuckets().add(bucket);
        }else{//bucket exists
            //just have to add this to the available and aggregate
            List<BucketFields> bucketFieldses = bucket.getFields();
            for ( BucketFields bucketFields : bucketFieldses){
                if (bucketFields.getName().compareToIgnoreCase("availableQty")== 0){
                    BigDecimal bigDecimal =  quantityFreePaid.getAvailable().add(new BigDecimal(bucketFields.getValue()));
                    bucketFields.setValue(String.valueOf(bigDecimal.intValue()));
                }
                //also need to update the times if we find a better range
                if (bucketFields.getName().compareToIgnoreCase("cuttoffDateTime")== 0){
                    //change it if needed to
                    if (bucketCutoffTime.isBefore(bucket.getBucketCutoffTime())){
                        bucket.setBucketCutoffTime(bucketCutoffTime);
                        bucketFields.setValue(availabilityServiceContextV4.calculateTimeBasedOnOffset(bucketCutoffTime, 0));
                    }
                }
                //if its the plu2 bucket the min and max fields would change
                if (bucketFields.getName().compareToIgnoreCase("pickupDateTimeMax")== 0) {
                    if (expectedMaxPickup.isAfter(bucket.getExpectedMaxPickup())){
                        bucket.setExpectedMaxPickup(expectedMaxPickup);
                        bucketFields.setValue(delivery.getDeliveryEstimate().getExpectationRangeMaxPickup());
                    }
                }
                if (bucketFields.getName().compareToIgnoreCase("deliverDateTimeMax")== 0) {
                    if (expectedMaxDeliver.isAfter(bucket.getExpectedMaxDeliver())){
                        bucket.setExpectedMaxDeliver(expectedMaxDeliver);
                        bucketFields.setValue(delivery.getDeliveryEstimate().getExpectationRangeMaxDeliver());
                    }
                }
                if (bucketName.contains("Plus2days") || bucketName.contains("Today") || bucketName.contains("Tomorrow")) {
                    if (bucketFields.getName().compareToIgnoreCase("deliverDateTimeMin")== 0) {
                        if (expectedMinDeliver.isBefore(bucket.getExpectedMinDeliver())){
                            bucket.setExpectedMinDeliver(expectedMinDeliver);
                            bucketFields.setValue(delivery.getDeliveryEstimate().getExpectationRangeMinDeliver());
                        }
                    }
                    if (bucketFields.getName().compareToIgnoreCase("pickupDateTimeMin")== 0) {
                        if (expectedMinPickup.isBefore(bucket.getExpectedMinPickup())){
                            bucket.setExpectedMinPickup(expectedMinPickup);
                            bucketFields.setValue(delivery.getDeliveryEstimate().getExpectationRangeMinPickup());
                        }
                    }
                }
            }
        }
    }

    public void createLocalBucket( Map<String, Bucket> bucketMap, QuantityFreePaid quantityFreePaid, AvailabilityServiceContextV4 availabilityServiceContextV4 ,BucketQuantity bucketQuantity){
        //this is a local store. if bucket does not exist create it
        //main counter primary DC and local store will contribute to this nucket
        Bucket bucket = bucketMap.get("Local Store");
        if (bucket==null){
            bucket= new Bucket();
            bucket.setType("Local Store");
            //first create the list for the fields
            List<BucketFields> bucketFieldses = new ArrayList<BucketFields>();

            //now create the fields for the bucket
            BucketFields bucketFields = new BucketFields();
            bucketFields.setName("availableQty");
            bucketFields.setValue(String.valueOf(quantityFreePaid.getAvailable().intValue()));
            bucketFieldses.add(bucketFields);


            //local bucket times are based on the store open
            BucketFields bucketFieldsPickup = new BucketFields();
            bucketFieldsPickup.setName("pickupDateTime");
            bucketFieldsPickup.setValue(availabilityServiceContextV4.calculateLocalTimeBasedOnTime(availabilityServiceContextV4.getStoreOpen(), availabilityServiceContextV4.getPickUpFromStoreOffset()));
            bucketFieldses.add(bucketFieldsPickup);

            BucketFields bucketFieldsDeliver = new BucketFields();
            bucketFieldsDeliver.setName("deliverDateTime");
            bucketFieldsDeliver.setValue(availabilityServiceContextV4.calculateLocalTimeBasedOnTime(availabilityServiceContextV4.getStoreOpen(), availabilityServiceContextV4.getDeliverFromStoreOffset()));
            bucketFieldses.add(bucketFieldsDeliver);

            BucketFields bucketFieldsOndemand = new BucketFields();
            bucketFieldsOndemand.setName("deliverOnDemandDateTime");
            bucketFieldsOndemand.setValue(availabilityServiceContextV4.calculateLocalTimeBasedOnTime(availabilityServiceContextV4.getStoreOpen(), availabilityServiceContextV4.getDeliverOnDemandOffset()));
            bucketFieldses.add(bucketFieldsOndemand);

            //set the list to the bucket
            bucket.setFields(bucketFieldses);
            bucketMap.put("Local Store", bucket);
            bucketQuantity.getBuckets().add(bucket);
        }else{
            /// if its a main counter we could have got an update
            //just have to add this to the available and aggregate
            List<BucketFields> bucketFieldses = bucket.getFields();
            for ( BucketFields bucketFields : bucketFieldses) {
                if (bucketFields.getName().compareToIgnoreCase("availableQty") == 0) {
                    BigDecimal bigDecimal = quantityFreePaid.getAvailable().add(new BigDecimal(bucketFields.getValue()));
                    bucketFields.setValue(String.valueOf(bigDecimal.intValue()));
                }
            }
        }
    }
}
