package com.genpt.nsight.v4.service.impl;

import com.genpt.nsight.NSightCache;
import com.genpt.nsight.v1.SplunkLogger;
import com.genpt.nsight.v1.model.*;
import com.genpt.nsight.v1.model.dto.SiteDTO;
import com.genpt.nsight.v4.AvailabilityServiceContextV4;
import com.genpt.nsight.v4.model.*;
import com.genpt.nsight.v4.model.DeliveryEstimateFreePaid;
import com.genpt.nsight.v4.model.DeliveryListFreePaid;
import com.genpt.nsight.v4.model.DetailResponseFreePaid;
import com.genpt.nsight.v4.model.PartFreePaid;
import com.genpt.nsight.v4.model.ProductAvailabilityResponseFreePaid;
import com.genpt.nsight.v4.model.SiteFreePaid;
import com.genpt.nsight.v4.service.ProductAvailabilityDetailServiceV4;
import com.genpt.nsight.v4.service.ProductAvailabilityFreePaidServiceV4;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.time.ZonedDateTime;
import java.util.*;

@Service("ProductAvailabilityFreePaidServiceV4")
public class ProductAvailabilityFreePaidServiceV4Impl implements ProductAvailabilityFreePaidServiceV4 {
    @Autowired
    NSightCache nSightCache;
    private static final Logger LOGGER = LoggerFactory.getLogger(ProductAvailabilityFreePaidServiceV4Impl.class);

    private String[] DayOfWeekColNames = new String[]{"Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"};

    private ProductAvailabilityDetailServiceV4 productAvailabilityDetailServiceV4;

    @Autowired
    public void setProductAvailabilityDetailServiceV2(ProductAvailabilityDetailServiceV4 productAvailabilityDetailServiceV4) {
        this.productAvailabilityDetailServiceV4 = productAvailabilityDetailServiceV4;
    }

    @Override
    public ProductAvailabilityResponseFreePaid getProductAvailability(AvailabilityServiceContextV4 availabilityServiceContextV3, boolean isVerification) {
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest());

        //set some defaults
        if (isVerification)
            availabilityServiceContextV3.getRequest().getActionRequest().getAction().setNeedDeliveryEstimateCutOffTimes(false);
        else
            availabilityServiceContextV3.getRequest().getActionRequest().getAction().setNeedDeliveryEstimateCutOffTimes(true);

        availabilityServiceContextV3.getRequest().getActionRequest().getAction().setNeedQuickestDelivery(false);

        LOGGER.debug(splunkLogger.toString()+ " request payload{} ",availabilityServiceContextV3.getRequest().toString() );
        long startTime = System.currentTimeMillis();

        String siteId = nSightCache.getAliasToSiteIdMap().get(availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getType() + "|" + availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getValue());
        //Since this is for verification, even inactive sites should be allowed
        if (siteId ==null && isVerification) {
            siteId = nSightCache.getInactiveAliasToSiteIdMap().get(availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getType()+ "|" +availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getValue());
        }

        SiteDTO callingStoreDTO = nSightCache.getSiteIdToSiteMap().get(siteId);

        if (!isVerification) {
            if (callingStoreDTO.getTimeZone() == null || callingStoreDTO.getTimeZone().equalsIgnoreCase("NAN")) {
                com.genpt.nsight.v4.model.ProductAvailabilityResponseFreePaid productAvailabilityResponseFreePaid = availabilityServiceContextV3.getProductAvailabilityResponseFreePaidTemplate(availabilityServiceContextV3);
                LOGGER.error(splunkLogger.toString() + " Invalid Data - Timezone is null in Nsight for requesting site: 3005");
                productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV3.createInvalidTimezoneTracking());
                return productAvailabilityResponseFreePaid;
            }
            if (nSightCache.getStoreHoursList().get(callingStoreDTO.getSiteIdentifier()) == null) {
                com.genpt.nsight.v4.model.ProductAvailabilityResponseFreePaid productAvailabilityResponseFreePaid = availabilityServiceContextV3.getProductAvailabilityResponseFreePaidTemplate(availabilityServiceContextV3);
                LOGGER.error(splunkLogger.toString() + " Invalid Data - No STore hours is null in Nsight for requesting site: 3005");
                productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV3.createInvalidStoreHoursTracking());
                return productAvailabilityResponseFreePaid;
            }
            if (callingStoreDTO.getTimeZoneID() == null) {
                com.genpt.nsight.v4.model.ProductAvailabilityResponseFreePaid productAvailabilityResponseFreePaid = availabilityServiceContextV3.getProductAvailabilityResponseFreePaidTemplate(availabilityServiceContextV3);
                LOGGER.error(splunkLogger.toString() + " Invalid Data - Timezone ID is null in NSight for requesting site: 3005");
                productAvailabilityResponseFreePaid.setTracking(availabilityServiceContextV3.createInvalidTimezoneIDTracking());
                return productAvailabilityResponseFreePaid;
            }
        }

/*        String storeTZ = callingStoreDTO.getTimeZone();
        ZoneOffset zoneOffSet = ZoneOffset.ofTotalSeconds((int) (Float.parseFloat(storeTZ) * 60 * 60));
        ZonedDateTime now = ZonedDateTime.now(ZoneId.of("UTC"));
        OffsetDateTime localTime = OffsetDateTime.now(zoneOffSet);

        boolean isStoresUTCNextDay = false;
        if (now.getDayOfMonth() != localTime.getDayOfMonth()) {
            isStoresUTCNextDay = true;
            // the UTC has passed into next day. need to get back a day into the stores date
            now = now.minus(1,ChronoUnit.DAYS);
        }
        if(availabilityServiceContextV3.getTimeOfDay() == 0L) {
            availabilityServiceContextV3.setDayOfWeekLocal(localTime.getDayOfWeek().getValue() - 1);
            availabilityServiceContextV3.setTimeOfDay(availabilityServiceContextV3.getMillisFromMidnight(now));
        }*/
        if(availabilityServiceContextV3.getTimeOfDay() == 0L && !isVerification) {
            availabilityServiceContextV3.setDayOfWeekLocalStore(callingStoreDTO);
        }

        //Get the detail response first
        com.genpt.nsight.v4.model.ProductAvailabilityResponse productAvailabilityResponse = productAvailabilityDetailServiceV4.getProductAvailability(availabilityServiceContextV3, isVerification);
        long buildResponseStartTime = System.currentTimeMillis();//start time for aggregate time taken
        com.genpt.nsight.v4.model.DetailResponse detailResponse = productAvailabilityResponse.getDetailResponse();//start creating the response

        //have a set to store sites that are processed to come up with unique list of sites
        Set<String> sites = new HashSet<String>();

        ProductAvailabilityResponseFreePaid productAvailabilityResponseFreePaid  = new ProductAvailabilityResponseFreePaid();
        productAvailabilityResponseFreePaid.setHeaderResponse(productAvailabilityResponse.getHeaderResponse());
        productAvailabilityResponseFreePaid.setCorrelationId(productAvailabilityResponse.getCorrelationId());
        productAvailabilityResponseFreePaid.setTracking(productAvailabilityResponse.getTracking());
        DetailResponseFreePaid detailResponseFreePaid = new DetailResponseFreePaid();
        if (availabilityServiceContextV3.getZipCodesSet().size() > 0) {
            com.genpt.nsight.v4.model.ShippingDestination  shippingDestination = new com.genpt.nsight.v4.model.ShippingDestination();
            shippingDestination.setZipcode(availabilityServiceContextV3.getZipCodesSet().iterator().next());
            detailResponseFreePaid.setShippingDestination(shippingDestination);
        }
        detailResponseFreePaid.setSites(new HashSet<SiteFreePaid>());
        detailResponseFreePaid.setParts(new ArrayList<PartFreePaid>());
        productAvailabilityResponseFreePaid.setDetailResponse(detailResponseFreePaid);

        Requestor requestor = availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor();//calling store

        //get the calling store siteID
        SiteDTO siteDTO = nSightCache.getSiteIdToSiteMap().get(siteId);

        //assign a boolean if calling store is a main counter
        boolean isCallingStoreMC = siteDTO.getMainCounterFlag();
        Set<String> combinedAllDCSet = new HashSet<String>();

        //iterate over all the parts requested
        detailResponse.getParts().forEach(part -> {
            if (true) {
                //we need a set where store the list of DCS that we are combining with main counters so we can remove them from the response at the end
                Set<String> combinedDCSet = new HashSet<String>();

                PartFreePaid partFreePaid = new PartFreePaid();
                Tracking trackingPart = new Tracking();
                trackingPart.setIsSuccess(part.getTracking().getIsSuccess());
                trackingPart.setLog(part.getTracking().getLog());
                trackingPart.setFault(part.getTracking().getFault());
                partFreePaid.setTracking(trackingPart);
                partFreePaid.setLineAbbrev(part.getLineAbbrev());
                partFreePaid.setPartNumber(part.getPartNumber());
                partFreePaid.setCatalogPartId(part.getCatalogPartId());
                partFreePaid.setSequenceNumber(part.getSequenceNumber());
                detailResponseFreePaid.getParts().add(partFreePaid);

                //Set the price
                if (part.getPrice() !=null)
                    partFreePaid.setPrice(part.getPrice());
                if (part.getSuperSedes()!=null)
                    partFreePaid.setSuperSedes(part.getSuperSedes());

                partFreePaid.setQuantities(new HashSet<QuantityFreePaid>());

                if (part.getSites() != null) {
                    // the part must have the local store in the list if we need to combine it with DC quantity
                    if (isCallingStoreMC  && nSightCache.getCombinedDCSet().contains(siteDTO.getPrimaryDCSiteId())){
                        boolean primaryDChasAvail = part.getSites().stream().filter(s -> s.getSiteId().equals(siteDTO.getPrimaryDCSiteId())).findFirst().isPresent();
                        boolean localMCHasAVail = part.getSites().stream().filter(s -> s.getSiteId().equals(siteDTO.getSiteIdentifier())).findFirst().isPresent();
                        if (primaryDChasAvail && !localMCHasAVail) {
                            //we need to add the LocalStore to the list. Need basic site details, dont need the RawDelivery for Local Store
                            //we are adding this just so that the subsequent logic for combining maincounter with DC works as expected
                            com.genpt.nsight.v4.model.Site adhocLocal = new com.genpt.nsight.v4.model.Site();
                            adhocLocal.setSiteIdentifiers(siteDTO.getSiteIdentifierList());
                            adhocLocal.setSiteId(siteDTO.getSiteIdentifier());
                            adhocLocal.setSiteType(siteDTO.getSiteType());
                            SiteRelationship localRelation = new SiteRelationship(null,"LocalStore");
                            List<SiteRelationship> localRelationList = new ArrayList<>();
                            localRelationList.add(localRelation);
                            adhocLocal.setSiteRelationships(localRelationList);
                            Quantity localQuantity = new Quantity();
                            adhocLocal.setQuantity(localQuantity);
                            adhocLocal.setAvailableQuantity(0);
                            Tracking tracking = new Tracking();
                            tracking.setIsSuccess(true);
                            adhocLocal.setTracking(tracking);
                            part.getSites().add(adhocLocal);
                        }
                    }//end of adding the local site if not present for combined inventory

                    part.getSites().forEach(site -> {// now iterate across all the sites for the part
                        SiteFreePaid siteFreePaid = new SiteFreePaid();

                        if (!sites.contains(site.getSiteId())) {// checking if we have already added this to the main Set of Sites in the response
                            sites.add(site.getSiteId());

                            siteFreePaid.setSiteCity(site.getSiteCity());
                            SiteDTO relatedSiteDTO = nSightCache.getSiteIdToSiteMap().get(site.getSiteId());
                            siteFreePaid.setSiteIdentifiers(relatedSiteDTO.getSiteIdentifierList());
                            siteFreePaid.setSiteRelationships(site.getSiteRelationships());
                            siteFreePaid.setSiteState(site.getSiteState());
                        	siteFreePaid.setSiteType(site.getSiteType());
                            if (nSightCache.getDisplayNameMap().get(site.getSiteId())!=null)
                                siteFreePaid.setDisplayName(nSightCache.getDisplayNameMap().get(site.getSiteId()));

                            if(relatedSiteDTO.getTimeZone()!=null && !relatedSiteDTO.getTimeZone().equalsIgnoreCase("NAN"))
                                siteFreePaid.setTracking(site.getTracking());
                            else {
                                Tracking tracking = new Tracking();
                                tracking.setIsSuccess(Boolean.FALSE);
                                tracking.setFault(Fault.builder().code("3005").message("Invalid Data - Timezone is null in Nsight for this site").build());
                                siteFreePaid.setTracking(tracking);
                                LOGGER.error(splunkLogger.toString() + " Invalid Data - Timezone is null in Nsight for this site: {}", relatedSiteDTO.getSiteIdentifier());
                            	
                            }
                            siteFreePaid.setSiteUUID(site.getSiteId());
                            //add to the hashset only after UUID is populated since thats being used for the hashcode
                            detailResponseFreePaid.getSites().add(siteFreePaid);

                            //add deliveries from this site
                            if (site.getRawDelivery() != null && site.getRawDelivery().size() > 0 && siteFreePaid.getTracking().getIsSuccess()) {
                                //add all the deliveries from the sorted map that has deliveries for each site
                                //there is one list based on the site to zip lookup and one based on the site to site lookup
                                List<com.genpt.nsight.v4.model.Delivery> deliverySorted = null;
                                if (availabilityServiceContextV3.isNeedSiteToSiteAndZip())
                                    deliverySorted = availabilityServiceContextV3.getDeliveryBySiteSorted().get(site.getSiteId());
                                else {
                                    if (availabilityServiceContextV3.getZipCodesSet().size() > 0)
                                        deliverySorted = availabilityServiceContextV3.getDeliveryByZipSorted().get(site.getSiteId());
                                    else
                                        deliverySorted = availabilityServiceContextV3.getDeliveryBySiteSorted().get(site.getSiteId());
                                }

                                DeliveryListFreePaid deliveryListFreePaid = new DeliveryListFreePaid();
                                deliveryListFreePaid.setDeliveryOptions(deliverySorted);
                                siteFreePaid.setDeliveryListFreePaid(deliveryListFreePaid);
                            }
                            //if its the Local store we need to add the deliveries based on current time and if its not for verification
                            if (site.getSiteId().compareToIgnoreCase(availabilityServiceContextV3.getRequestingSite().getSiteIdentifier()) == 0 && !isVerification) {
                                createLocalDeliveries(availabilityServiceContextV3, siteFreePaid);
                                //done with localstore deleveries
                            }
                        }// we have created the site with it deliveries and added it to the Site set for this part

                        //now add the quantity for this site
                        QuantityFreePaid quantityFreePaid = new QuantityFreePaid();
                        //adding logic on 03312020 to sum up main counter stores for certain DC's needed due to COVID-19(total cases 875K)
                        SiteDTO currentSiteDTO = nSightCache.getSiteIdToSiteMap().get(site.getSiteId());
                        if (currentSiteDTO.getMainCounterFlag()){
                            //initialize with the value
                            if(site.getQuantity()!=null && site.getQuantity().getAvailable()!=null)
                                quantityFreePaid.setAvailable(site.getQuantity().getAvailable());

                            //get this sites primary DC
                            String dcSiteID = currentSiteDTO.getPrimaryDCSiteId();
                            if (dcSiteID!=null && nSightCache.getCombinedDCSet().contains(dcSiteID)){// check if the DC exists in the list of site
                                boolean primaryDChasAvail = part.getSites().stream().filter(s -> s.getSiteId().equals(dcSiteID)).findFirst().isPresent();
                                if (primaryDChasAvail){
                                    Optional<com.genpt.nsight.v4.model.Site> dcSite  =  part.getSites().stream().filter(s -> s.getSiteId().equals(dcSiteID)).findFirst();
                                    if (dcSite!=null && dcSite.isPresent()) {
                                        // add combined inventory
                                        if(dcSite.get().getQuantity()!=null && dcSite.get().getQuantity().getAvailable()!=null)
                                            quantityFreePaid.setAvailable(quantityFreePaid.getAvailable().add(dcSite.get().getQuantity().getAvailable()));                                        combinedDCSet.add(dcSiteID);
                                        if (!combinedAllDCSet.contains(dcSiteID))
                                            combinedAllDCSet.add(dcSiteID);
                                    }
                                }
                            }
                        }else {
                            if(site.getQuantity()!=null && site.getQuantity().getAvailable()!=null)
                                quantityFreePaid.setAvailable(site.getQuantity().getAvailable());
                        }
                        // add the following additional attributes for inventory verification
                        if (site.getQuantity()!=null) {
                            if (site.getQuantity().getOnHand() !=null)
                                quantityFreePaid.setOnHand(site.getQuantity().getOnHand());
                            if (site.getQuantity().getInTransit()!= null)
                                quantityFreePaid.setInTransit(site.getQuantity().getInTransit());
                            if (site.getQuantity().getOnOrder()!=null)
                                quantityFreePaid.setOnOrder(site.getQuantity().getOnOrder());
                            if (site.getQuantity().getOnOrderDate()!=null)
                                quantityFreePaid.setOnOrderDate(site.getQuantity().getOnOrderDate());
                            if (site.getQuantity().getInTransitDate()!=null)
                                quantityFreePaid.setInTransitDate(site.getQuantity().getInTransitDate());
                            if (site.getSourceSystemDataAsOfDate()!=null)
                                quantityFreePaid.setSourceSystemDataAsOfDate(site.getSourceSystemDataAsOfDate());
                        }

                        //this combined availability is only for the combined friends site relationship. this not related to combining main counters
                        if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedCombinedInventory()){
                            if (site.getQuantity() != null && site.getQuantity().getCombinedAvailability()!=null)
                                quantityFreePaid.setCombinedAvailability(site.getQuantity().getCombinedAvailability());
                        }

                        quantityFreePaid.setSiteUUID(site.getSiteId());

                        //add availability details if asked for in the request
                        if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedAvailabilityDetails()) {
                            if (site.getQuantity() !=null && site.getQuantity().getIncrementalSellQuantity()!=null)
                                quantityFreePaid.setIncrementalSellQuantity(site.getQuantity().getIncrementalSellQuantity());
                            if (site.getQuantity() != null && site.getQuantity().getMaxStockedQuantity() != null && site.getQuantity().getMaxStockedQuantity() != Integer.MIN_VALUE)
                                quantityFreePaid.setMaxStockedQuantity(site.getQuantity().getMaxStockedQuantity());
                            if (site.getQuantity() != null && site.getQuantity().getMinimumSellQuantity()!=null)
                                quantityFreePaid.setMinimumSellQuantity(site.getQuantity().getMinimumSellQuantity());
                            if (site.getQuantity() != null && site.getQuantity().getMinStockedQuantity() != null && site.getQuantity().getMinStockedQuantity() != Integer.MIN_VALUE)
                                quantityFreePaid.setMinStockedQuantity(site.getQuantity().getMinStockedQuantity());
                            if (site.getPartAttributes() != null) {
                                PartAttributes partAttributes = new PartAttributes();
                                partAttributes.setPartDescription(site.getPartAttributes().getPartDescription());
                                partAttributes.setUnitOfMeasure(site.getPartAttributes().getUnitOfMeasure());
                                quantityFreePaid.setPartAttributes(partAttributes);
                            }
                        }
                        //for suppliers we need to factor in the order since last update quantity into available
                        if (site.getSiteType()!=null && site.getSiteType().compareToIgnoreCase("Supplier")==0 && quantityFreePaid.getAvailable()!=null && site.getQuantity()!=null && site.getQuantity().getOrderSinceLastUpdate() !=null && site.getQuantity().getOrderSinceLastUpdate() != Integer.MIN_VALUE){
                            quantityFreePaid.setAvailable(quantityFreePaid.getAvailable().subtract(BigDecimal.valueOf(site.getQuantity().getOrderSinceLastUpdate())));
                            if (quantityFreePaid.getAvailable().compareTo(new BigDecimal(0))==-1)
                                quantityFreePaid.setAvailable(new BigDecimal(0));
                        }
                        partFreePaid.getQuantities().add(quantityFreePaid);
                    });//for each site within a part
                }//valid sites array present

                if (part.getProductKit() != null) {
                    partFreePaid.setProductKit(part.getProductKit()); //This is a virtual part
                }
                //at this point we can remove the DC's from the response if combined with main counters
                if (combinedDCSet.size() > 0){
                    for (String combinedDCSiteID : combinedDCSet) {
                        partFreePaid.getQuantities().removeIf(quantityFreePaid -> quantityFreePaid.getSiteUUID().equals(combinedDCSiteID));
                    }
                }
            }// done for a valid part. end of if not null

        });//for each part end

        // at this point should we also remove the DC from the sites list if it was combined?
        if (combinedAllDCSet.size() > 0){
            for (String combinedAllDCSetID : combinedAllDCSet) {
                detailResponseFreePaid.getSites().removeIf(siteFreePaid -> siteFreePaid.getSiteUUID().equals(combinedAllDCSetID));
            }
        }
        int totalPartsRequested = getTotalPartsRequested(availabilityServiceContextV3);
        LOGGER.debug(splunkLogger.toString() + " Time taken to build response = {} and the total parts requested are = {} ",  (System.currentTimeMillis() - buildResponseStartTime), totalPartsRequested);
        LOGGER.info(splunkLogger.toString() + " Total time within FreePaid MicroService = {} and the total parts requested are = {} ", (System.currentTimeMillis() - startTime), totalPartsRequested);

        return productAvailabilityResponseFreePaid;
    }

    private static int getTotalPartsRequested(AvailabilityServiceContextV4 availabilityServiceContextV4) {
        if(availabilityServiceContextV4 != null && availabilityServiceContextV4.getRequest() != null
                && availabilityServiceContextV4.getRequest().getDetailRequest() != null && availabilityServiceContextV4.getRequest().getDetailRequest().getParts() != null) {
            return availabilityServiceContextV4.getRequest().getDetailRequest().getParts().size();
        }
        return 0;
    }

    private void createLocalDeliveries(AvailabilityServiceContextV4 availabilityServiceContextV3, SiteFreePaid siteFreePaid){
        DeliveryListFreePaid deliveryListFreePaid = new DeliveryListFreePaid();
        ArrayList<com.genpt.nsight.v4.model.Delivery> deliveries =  new ArrayList<>();
        deliveryListFreePaid.setDeliveryOptions(deliveries);
        com.genpt.nsight.v4.model.Delivery delivery = new Delivery();
        //populate the delivery which depends on store hours
        delivery.setCutOffDateTime(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), 0));
        delivery.setQuickestOption(true);
        delivery.setFreeDelivery(true);
        delivery.setDeliveryArrivalSeq(1);
        delivery.setDeliveryType("LocalStore");
        // also need to populate the day of the week based on when available
        delivery.setDayofweekString(DayOfWeekColNames[availabilityServiceContextV3.getStoreOpen().getDayOfWeek().getValue()-1]);

        DeliveryEstimateFreePaid deliveryEstimateFreePaid=  new DeliveryEstimateFreePaid();
        deliveryEstimateFreePaid.setDurationDesignator("Minutes");
        //populate the new expectations fields
        populateLocalExpectations( availabilityServiceContextV3, deliveryEstimateFreePaid);

        delivery.setDeliveryEstimate(deliveryEstimateFreePaid);
        //delivery.setCutOffDateTime();
        deliveries.add(delivery);
        siteFreePaid.setDeliveryListFreePaid(deliveryListFreePaid);
    }

    private void populateLocalExpectations(AvailabilityServiceContextV4 availabilityServiceContextV3, DeliveryEstimateFreePaid deliveryEstimateFreePaid){
        ArrayList<Expectation> expectationArray =  new ArrayList<>();
        //we need to create only expectations that are passed in
//        if (availabilityServiceContextV3.getPickUpFromStoreOffset()!=0) {
        if (availabilityServiceContextV3.getPickUpFromStoreOffsetRaw()!=null) {
            Expectation expectation = new Expectation();
            int offsetPickup = availabilityServiceContextV3.getPickUpFromStoreOffset();
            expectation.setExpectationRangeMax(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), offsetPickup));
            expectation.setExpectationRangeMin(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), offsetPickup));
            expectation.setOffsetName("Pickup from Store");
            expectation.setOffsetValue(String.valueOf(availabilityServiceContextV3.getPickUpFromStoreOffset()));
            deliveryEstimateFreePaid.setExpectationRangeMaxPickup(expectation.getExpectationRangeMax());
            deliveryEstimateFreePaid.setExpectationRangeMinPickup(expectation.getExpectationRangeMin());
            expectationArray.add(expectation);
        }
        if (availabilityServiceContextV3.getDeliverFromStoreOffsetRaw()!=null) {
            Expectation expectationDeliver = new Expectation();
            int offsetDeliver = availabilityServiceContextV3.getDeliverFromStoreOffset();
            expectationDeliver.setExpectationRangeMax(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), offsetDeliver));
            expectationDeliver.setExpectationRangeMin(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), offsetDeliver));
            expectationDeliver.setOffsetName("Deliver from Store");
            expectationDeliver.setOffsetValue(String.valueOf(offsetDeliver));
            deliveryEstimateFreePaid.setExpectationRangeMaxDeliver(expectationDeliver.getExpectationRangeMax());
            deliveryEstimateFreePaid.setExpectationRangeMinDeliver(expectationDeliver.getExpectationRangeMin());
            expectationArray.add(expectationDeliver);
        }

        if (availabilityServiceContextV3.getDeliverOnDemandOffsetRaw()!=null) {
            Expectation expectationDeliveronDemand = new Expectation();
            int offsetDeliveronDemand = availabilityServiceContextV3.getDeliverOnDemandOffset();
            expectationDeliveronDemand.setExpectationRangeMax(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), offsetDeliveronDemand));
            expectationDeliveronDemand.setExpectationRangeMin(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), offsetDeliveronDemand));
            expectationDeliveronDemand.setOffsetName("Deliver on Demand");
            expectationDeliveronDemand.setOffsetValue(String.valueOf(offsetDeliveronDemand));
            deliveryEstimateFreePaid.setExpectationRangeMaxOnDemand(expectationDeliveronDemand.getExpectationRangeMax());
            deliveryEstimateFreePaid.setExpectationRangeMinOnDemand(expectationDeliveronDemand.getExpectationRangeMin());
            expectationArray.add(expectationDeliveronDemand);
        }
        if (expectationArray !=null && expectationArray.size() == 0){
            //populate default
            Expectation expectation = new Expectation();
            expectation.setExpectationRangeMax(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), 0));
            expectation.setExpectationRangeMin(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), 0));
            expectationArray.add(expectation);
        }

        deliveryEstimateFreePaid.setExpectations(expectationArray);

        //we need to set which is the quickest expectation at this point
        if (expectationArray!=null && expectationArray.size() > 0){
            for (Expectation expectation : expectationArray){
                ZonedDateTime curExpectedMax = null;
                ZonedDateTime curExpectedMin  = null;
                curExpectedMin = availabilityServiceContextV3.parseDateTime(expectation.getExpectationRangeMin(), false);
                curExpectedMax = availabilityServiceContextV3.parseDateTime(expectation.getExpectationRangeMax(), false);

                if (deliveryEstimateFreePaid.getQuickestExpectationOfAllMin()==null) {
                    deliveryEstimateFreePaid.setQuickestExpectationOfAllMin(curExpectedMin);
                    deliveryEstimateFreePaid.setQuickestExpectationOfAllMax(curExpectedMax);
                }
                else{
                    if (curExpectedMin.isBefore(deliveryEstimateFreePaid.getQuickestExpectationOfAllMin())) {
                        deliveryEstimateFreePaid.setQuickestExpectationOfAllMin(curExpectedMin);
                        deliveryEstimateFreePaid.setQuickestExpectationOfAllMax(curExpectedMax);
                    }
                }
            }
        }else{
            //we can default to the expected minimum since no offsets were passed in
            ZonedDateTime curExpectedMin = null;
            ZonedDateTime curExpectedMax = null;
            curExpectedMin = availabilityServiceContextV3.parseDateTime(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), 0), false);
            curExpectedMax = availabilityServiceContextV3.parseDateTime(availabilityServiceContextV3.calculateLocalTimeBasedOnTime(availabilityServiceContextV3.getStoreOpen(), 0), false);

            deliveryEstimateFreePaid.setQuickestExpectationOfAllMin(curExpectedMin);
            deliveryEstimateFreePaid.setQuickestExpectationOfAllMax(curExpectedMax);
        }
    }

/*    protected long getMillisFromMidnight(ZonedDateTime now) {
        ZonedDateTime midnight = now.truncatedTo(ChronoUnit.DAYS);
        Duration duration = Duration.between(midnight, now);
        TimeZone tz  =  TimeZone.getDefault();

        boolean isOn = ZoneId.of(tz.getID())  // Represent a specific time zone, the history of past, present, and future changes to the offset-from-UTC used by the people of a certain region.
                .getRules()                // Obtain the list of those changes in offset.
                .isDaylightSavings(        // See if the people of this region are observing Daylight Saving Time at a specific moment.
                        Instant.now()          // Specify the moment. Here we capture the current moment at runtime.
                )    ;                      // Returns a boolean.

        if (isOn)
            return duration.getSeconds() * 1000L +tz.getDSTSavings();
        else
            return duration.getSeconds()  * 1000L ;
    }

    private String getDateTime(int days, long timeInMills, ZonedDateTime now, boolean nextDay) {
        long minutes = timeInMills / (60 * 1000L);
        String dateTime = null;
        try {
            if (nextDay)
                dateTime = DateTimeFormatter.ISO_INSTANT.format(now.plusDays(days+1).plusMinutes(minutes));
            else
                dateTime = DateTimeFormatter.ISO_INSTANT.format(now.plusDays(days).plusMinutes(minutes));
        } catch (Exception e) {
            e.printStackTrace();
        }
        return dateTime;
    }*/
}
