package com.genpt.nsight.v4.service.impl;

import com.genpt.nsight.NSightCache;
import com.genpt.nsight.config.NSightProperties;
import com.genpt.nsight.v1.model.*;
import com.genpt.nsight.v2.model.DetailResponse;
import com.genpt.nsight.v2.model.ProductAvailabilityResponse;
import com.genpt.nsight.v4.AvailabilityServiceContextV4;
import com.genpt.nsight.v4.SplunkLogger;
import com.genpt.nsight.v4.model.PartFreePaid;
import com.genpt.nsight.v4.model.ProductAvailabilityResponseFreePaid;
import com.genpt.nsight.v4.model.SiteFreePaid;
import com.genpt.nsight.v4.service.ProductAvailabilityFreePaidServiceV4;
import com.genpt.nsight.v4.service.ProductAvailabilityVerificationV4;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

@Service("ProductAvailabilityVerificationV4")
public class ProductAvailabilityVerificationV4Impl implements ProductAvailabilityVerificationV4 {

    @Autowired
    NSightCache nSightCache;
    private static final Logger LOGGER = LoggerFactory.getLogger(ProductAvailabilityVerificationV4Impl.class);

    @Autowired
    NSightProperties nSightProperties;

    private ProductAvailabilityFreePaidServiceV4 productAvailabilityFreePaidServiceV4;

    @Autowired
    public void setProductAvailabilityFreePaidServiceV4(ProductAvailabilityFreePaidServiceV4 productAvailabilityFreePaidServiceV4) {
        this.productAvailabilityFreePaidServiceV4 = productAvailabilityFreePaidServiceV4;
    }

    @Override
    public ProductAvailabilityResponse getProductAvailability(AvailabilityServiceContextV4 availabilityServiceContextV4) {
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV4.getRequest().getHeaderRequest());
        LOGGER.debug(splunkLogger.toString() + " Received Inventory Verification request v4 for store" );

        long startTime = System.currentTimeMillis();
        ProductAvailabilityResponse productAvailabilityResponse = new ProductAvailabilityResponse();
        productAvailabilityResponse.setCorrelationId(availabilityServiceContextV4.getCorrelationID());
        HeaderResponse headerResponse = new HeaderResponse();
        headerResponse.setCountryCode(availabilityServiceContextV4.getRequest().getHeaderRequest().getCountryCode());
        headerResponse.setLanguageCode(availabilityServiceContextV4.getRequest().getHeaderRequest().getLanguageCode());
        headerResponse.setRequestor(availabilityServiceContextV4.getRequest().getHeaderRequest().getRequestor());
        headerResponse.setResponder(availabilityServiceContextV4.getRequest().getHeaderRequest().getResponder());
        headerResponse.setTimestamp(Long.valueOf(availabilityServiceContextV4.getRequest().getHeaderRequest().getTimestamp()));
        productAvailabilityResponse.setHeaderResponse(headerResponse);
        productAvailabilityResponse.setTracking(new Tracking());
        productAvailabilityResponse.getTracking().setIsSuccess(true);
        //call the V4 free paid service
        ProductAvailabilityResponseFreePaid productAvailabilityResponseFreePaid = productAvailabilityFreePaidServiceV4.getProductAvailability(availabilityServiceContextV4, true);
        if (productAvailabilityResponseFreePaid.getTracking().getLog()!=null)
            productAvailabilityResponse.getTracking().setLog(productAvailabilityResponseFreePaid.getTracking().getLog());
        if (productAvailabilityResponseFreePaid.getDetailResponse() != null) {
            //store the site details for each site in a hashmap
            Map<String, SiteFreePaid> siteFreePaidMap = new HashMap<>();
            for (SiteFreePaid siteFreePaid : productAvailabilityResponseFreePaid.getDetailResponse().getSites()) {
                siteFreePaidMap.put(siteFreePaid.getSiteUUID(), siteFreePaid);
            }
            // now transform response from v4 to v2
            productAvailabilityResponse.setDetailResponse(new DetailResponse());
            productAvailabilityResponse.getDetailResponse().setParts(new ArrayList<>());
            for (PartFreePaid partFreePaid : productAvailabilityResponseFreePaid.getDetailResponse().getParts()) {
                Part part = new Part();
                part.setLineAbbrev(partFreePaid.getLineAbbrev());
                part.setPartNumber(partFreePaid.getPartNumber());
                if (partFreePaid.getQuantities().size() > 0) {
                    part.setTracking(new Tracking());
                    part.getTracking().setIsSuccess(true);

                    part.setSites(new HashSet<>());
                    for (QuantityFreePaid quantityFreePaid : partFreePaid.getQuantities()) {// this is one per site
                        //Create a new site
                        Site site = new Site();
                        SiteFreePaid siteFreePaid = siteFreePaidMap.get(quantityFreePaid.getSiteUUID());
                        site.setSiteType(siteFreePaid.getSiteType());
                        site.setSiteIdentifiers(siteFreePaid.getSiteIdentifiers());
                        site.setSiteRelationships(siteFreePaid.getSiteRelationships());
                        site.setSourceSystemDataAsOfDate(quantityFreePaid.getSourceSystemDataAsOfDate());
                        Quantity quantity = new Quantity();
                        quantity.setAvailable(quantityFreePaid.getAvailable());
                        //since onhand is mandatory on the TAMS end, we will default if there is no value
                        if(quantityFreePaid.getOnHand()!=null)
                            quantity.setOnHand(quantityFreePaid.getOnHand());
                        else
                            quantity.setOnHand(quantityFreePaid.getAvailable());
                        quantity.setInTransit(quantityFreePaid.getInTransit());
                        quantity.setInTransitDate(quantityFreePaid.getInTransitDate());
                        quantity.setOnOrder(quantityFreePaid.getOnOrder());
                        quantity.setOnOrderDate(quantityFreePaid.getOnOrderDate());
                        quantity.setMinStockedQuantity(quantityFreePaid.getMinStockedQuantity());
                        quantity.setMaxStockedQuantity(quantityFreePaid.getMaxStockedQuantity());
                        quantity.setMinimumSellQuantity(quantityFreePaid.getMinimumSellQuantity());
                        quantity.setIncrementalSellQuantity(quantityFreePaid.getIncrementalSellQuantity());
                        site.setPartAttributes(quantityFreePaid.getPartAttributes());
                        site.setQuantity(quantity);
                        site.setTracking(new Tracking());
                        site.getTracking().setIsSuccess(true);
                        part.getSites().add(site);
                    }
                } else {
                    part.setTracking(new Tracking());
                    part.getTracking().setIsSuccess(true);
                    part.getTracking().setLog(new ArrayList<>());
                    part.getTracking().getLog().add(new Log("2001", "Part Not Found"));
                }
                productAvailabilityResponse.getDetailResponse().getParts().add(part);
            }
        }else{
            productAvailabilityResponse.getTracking().setIsSuccess(false);
            if (productAvailabilityResponseFreePaid.getTracking().getFault()!=null)
                productAvailabilityResponse.getTracking().setFault(productAvailabilityResponseFreePaid.getTracking().getFault());
            if (productAvailabilityResponseFreePaid.getTracking().getLog()!=null)
                productAvailabilityResponse.getTracking().setLog(productAvailabilityResponseFreePaid.getTracking().getLog());
        }
        return productAvailabilityResponse;
    }
}
