package com.genpt.nsight.v4.service.impl;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.genpt.nsight.config.IbizProductConfig;
import com.genpt.nsight.v4.model.majoraccounts.ProductTranslateRequest;
import com.genpt.nsight.v4.model.majoraccounts.ProductTranslateResponse;
import com.genpt.nsight.v4.service.ProductTranslateService;
import org.apache.http.impl.client.CloseableHttpClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.client.support.BasicAuthorizationInterceptor;
import org.springframework.stereotype.Service;
import com.genpt.common.NFuseSSLHttpConnection;
import org.springframework.web.client.RestTemplate;

import java.util.Arrays;

@Service("ProductTranslateService")
public class ProductTranslateServiceImpl implements ProductTranslateService {

    @Autowired
    IbizProductConfig ibizProductConfig;

    private static final Logger LOGGER = LoggerFactory.getLogger(ProductTranslateServiceImpl.class);

    @Override
    public ProductTranslateResponse translateProducts(ProductTranslateRequest productTranslateRequest) {
        long time = System.currentTimeMillis();
        String username = ibizProductConfig.getUsername();
        String password = ibizProductConfig.getPassword();
        String finalURL = ibizProductConfig.getUrl();
        CloseableHttpClient httpClient = null;
        ProductTranslateResponse productTranslateResponse = null;

        try {
            NFuseSSLHttpConnection nFuseSSLHttpConnection = new NFuseSSLHttpConnection();
            httpClient = nFuseSSLHttpConnection.getTrustAllHttpsClient();
            HttpComponentsClientHttpRequestFactory factory = new HttpComponentsClientHttpRequestFactory();
            factory.setHttpClient(httpClient);
            HttpHeaders headers = new HttpHeaders();
            headers.setContentType(MediaType.APPLICATION_JSON);
            headers.setAccept(Arrays.asList(MediaType.APPLICATION_JSON));
            factory.setReadTimeout(ibizProductConfig.getTimeout());
            factory.setConnectTimeout(ibizProductConfig.getTimeout() / 2);
            RestTemplate restTemplate = new RestTemplate(factory);
            //restTemplate.getInterceptors().add(new BasicAuthorizationInterceptor(username, password));
            ObjectMapper objectMapper = new ObjectMapper();
            objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
            objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
            String jsonStr = objectMapper.writeValueAsString(productTranslateRequest);
            HttpEntity<String> request = new HttpEntity<String>(jsonStr, headers);
            LOGGER.info("Ibiz Product Translate Service  URL: {} ,  Request: {} , Timeout: {} ", finalURL, jsonStr, ibizProductConfig.getTimeout());
            String response = restTemplate.postForObject(finalURL, request, String.class);
            LOGGER.info("Ibiz Product Translate Service  URL: {} ,  Response: {} , Timeout: {} , Time Taken {} ms ", finalURL, response, ibizProductConfig.getTimeout(), (System.currentTimeMillis() - time));

            productTranslateResponse = objectMapper.readValue(response, ProductTranslateResponse.class);

        }
        catch(Exception ex){
            LOGGER.error(ex.getMessage());
        }
        return productTranslateResponse;
    }
}
