package com.genpt.nsight.v4.service.impl;

import com.genpt.nsight.NSightCache;
import com.genpt.nsight.v1.model.*;
import com.genpt.nsight.v1.model.dto.SiteDTO;
import com.genpt.nsight.v1.model.dto.SiteRelationDTO;
import com.genpt.nsight.v2.SplunkLogger;
import com.genpt.nsight.v3.service.AvailabilityServiceContextV3;
import com.genpt.nsight.v4.AvailabilityServiceContextV4;
import com.genpt.nsight.v4.service.SiteRelationshipServiceV4;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

@Service("SiteRelationshipServiceV4")
public class SiteRelationshipServiceV4Impl implements SiteRelationshipServiceV4 {

    private static final Logger LOGGER = LoggerFactory.getLogger(SiteRelationshipServiceV4Impl.class);

    @Autowired
    NSightCache nSightCache;

    @Override
    public void setRelatedSites(AvailabilityServiceContextV4 availabilityServiceContextV3) {

        //Populate all the related sites that we would querying inventory for
        com.genpt.nsight.v4.model.DetailRequest detailRequest = availabilityServiceContextV3.getRequest().getDetailRequest();
        Requestor requestor = availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor();

        if (detailRequest.getInventoryRelationshipsRequested().isLocalStore()) {
            RelatedSite relatedSite = new RelatedSite();
            relatedSite.setPartKey(availabilityServiceContextV3.getRequestingSite().getStorePartitionKey());
            relatedSite.setSiteId(availabilityServiceContextV3.getRequestingSite().getSiteIdentifier());
            relatedSite.setRelationType("LocalStore");
            relatedSite.setType(availabilityServiceContextV3.getRequestingSite().getSiteType());
            //pos attributes
            relatedSite.setPosType(availabilityServiceContextV3.getRequestingSite().getPosType());
            relatedSite.setPosURL(availabilityServiceContextV3.getRequestingSite().getPosURL());
            relatedSite.setPosPassword(availabilityServiceContextV3.getRequestingSite().getPosPassword());
            relatedSite.setRetrieveFromStore(availabilityServiceContextV3.getRequestingSite().getRetrieveFromStore());

            addSiteRelation(availabilityServiceContextV3.getRequestingSite().getSiteIdentifier(), "LocalStore", null, availabilityServiceContextV3.getSiteRelationMap());
            availabilityServiceContextV3.getRelatedSites().add(relatedSite);
        }

        if (detailRequest.getInventoryRelationshipsRequested().isLocalStoreOwnershipGroup() && !"JDE".equalsIgnoreCase(requestor.getRequestorApplication())) {
            List<SiteRelationDTO> siteRelationDTOList = availabilityServiceContextV3.getRequestingSite().getSiteRelationDTOList();
            if (siteRelationDTOList != null && siteRelationDTOList.size() > 0) {
                siteRelationDTOList.stream().filter(siteRelationDTO -> siteRelationDTO.getRelationshipType().equalsIgnoreCase("Ownership Group") && nSightCache.getSiteIdToSiteMap().get(siteRelationDTO.getRelatedSiteIdentifier()).getIsActive())
                        .map((siteRelationDTO) -> {
                            RelatedSite relatedSite = new RelatedSite();
                            relatedSite.setPartKey(siteRelationDTO.getRelatedStorePartitionKey());
                            relatedSite.setSiteId(siteRelationDTO.getRelatedSiteIdentifier());
                            relatedSite.setRelationType("Ownership Group");
                            relatedSite.setType(siteRelationDTO.getRelatedSiteType());
                            relatedSite.setPosType(siteRelationDTO.getPosType());
                            relatedSite.setPosURL(siteRelationDTO.getPosURL());
                            relatedSite.setPosPassword(siteRelationDTO.getPosPassword());

                            addSiteRelation(siteRelationDTO.getRelatedSiteIdentifier(), siteRelationDTO.getRelationshipType(), siteRelationDTO.getRelationshipRank(), availabilityServiceContextV3.getSiteRelationMap());
                            return relatedSite;
                        }).forEach(availabilityServiceContextV3.getRelatedSites()::add);
            }
        }

        OtherRelationShips otherRelationShips = detailRequest.getInventoryRelationshipsRequested().getOtherRelationShips();
        if (otherRelationShips != null && otherRelationShips.getRelationshipName() != null && otherRelationShips.getRelationshipName().size() > 0) {
            List<SiteRelationDTO> siteRelationDTOList = availabilityServiceContextV3.getRequestingSite().getSiteRelationDTOList();
            if (siteRelationDTOList != null && siteRelationDTOList.size() > 0) {
                siteRelationDTOList.stream().filter(siteRelationDTO -> otherRelationShips.getRelationshipName().contains(siteRelationDTO.getRelationshipType()) && nSightCache.getSiteIdToSiteMap().get(siteRelationDTO.getRelatedSiteIdentifier()).getIsActive())
                        .map((siteRelationDTO) -> {
                            RelatedSite relatedSite = new RelatedSite();
                            relatedSite.setPartKey(siteRelationDTO.getRelatedStorePartitionKey());
                            relatedSite.setSiteId(siteRelationDTO.getRelatedSiteIdentifier());
                            relatedSite.setRelationType(siteRelationDTO.getRelationshipType());
                            relatedSite.setType(siteRelationDTO.getRelatedSiteType());
                            relatedSite.setPosType(siteRelationDTO.getPosType());
                            relatedSite.setPosURL(siteRelationDTO.getPosURL());
                            relatedSite.setPosPassword(siteRelationDTO.getPosPassword());

                            addSiteRelation(siteRelationDTO.getRelatedSiteIdentifier(), siteRelationDTO.getRelationshipType(), siteRelationDTO.getRelationshipRank(), availabilityServiceContextV3.getSiteRelationMap());
                            return relatedSite;
                        }).forEach(availabilityServiceContextV3.getRelatedSites()::add);
            }
        }

        // Add Other sites if present in the request. Since there wont be a relationship for these,
        // add Othersites are the relationship between the two.
        OtherSites otherSites = detailRequest.getInventoryRelationshipsRequested().getOtherSites();
        if (otherSites != null) {
            otherSites.getSiteIdentifiers().stream().map(siteIdentifier -> {
                String siteId = nSightCache.getAliasToSiteIdMap().get(siteIdentifier.getName() + "|" + siteIdentifier.getValue());
                RelatedSite relatedSite = null;
                if (siteId != null &&  nSightCache.getSiteIdToSiteMap().get(siteId).getIsActive()) {
                        SiteDTO _siteDTO = nSightCache.getSiteIdToSiteMap().get(siteId);
                        relatedSite = new RelatedSite();
                        relatedSite.setSiteId(_siteDTO.getSiteIdentifier());
                        relatedSite.setPartKey(_siteDTO.getStorePartitionKey());
                        relatedSite.setType(_siteDTO.getSiteType());
                        relatedSite.setRelationType("OtherSites");
                        relatedSite.setPosType(_siteDTO.getPosType());
                        relatedSite.setPosURL(_siteDTO.getPosURL());
                        relatedSite.setPosPassword(_siteDTO.getPosPassword());

                    ///probably need to add this new relationship type to the relationship map
                        addSiteRelation(_siteDTO.getSiteIdentifier(),"OtherSites", 99, availabilityServiceContextV3.getSiteRelationMap());

                            // If its a supplier or a DC/not Primary DC, then add the AllSuppliers relationship
                            if (_siteDTO.getSiteType().equalsIgnoreCase("Supplier") || _siteDTO.getSiteType().equalsIgnoreCase("DC")){
                                relatedSite = new RelatedSite();
                                relatedSite.setSiteId(_siteDTO.getSiteIdentifier());
                                relatedSite.setPartKey(_siteDTO.getStorePartitionKey());
                                relatedSite.setType(_siteDTO.getSiteType());

                                if (_siteDTO.getSiteType().equalsIgnoreCase("Supplier") && availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllSuppliers()) {
                                    relatedSite.setRelationType("AllSuppliers");
                                    addSiteRelation(_siteDTO.getSiteIdentifier(),relatedSite.getRelationType(), 99, availabilityServiceContextV3.getSiteRelationMap());
                                }
                                if(_siteDTO.getSiteType().equalsIgnoreCase("DC") && availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllOtherDCs()) {
                                    if (availabilityServiceContextV3.getPrimaryDC()!= null) {
                                        if (!availabilityServiceContextV3.getPrimaryDC().getSiteId().equalsIgnoreCase(_siteDTO.getSiteIdentifier())) {
                                            relatedSite.setRelationType("AllOtherDCs");
                                            addSiteRelation(_siteDTO.getSiteIdentifier(), relatedSite.getRelationType(), 99, availabilityServiceContextV3.getSiteRelationMap());
                                        }
                                    }
                                }
                            }
                        }
                        return relatedSite;
                    }).filter(rs -> rs != null).forEach(availabilityServiceContextV3.getRelatedSites()::add);
                }
            }

    @Override
    public void setPrimaryDC(AvailabilityServiceContextV4 availabilityServiceContextV3) {
        com.genpt.nsight.v4.model.DetailRequest detailRequest = availabilityServiceContextV3.getRequest().getDetailRequest();

        RelatedSite primaryDC = null;

        //if (detailRequest.getInventoryRelationshipsRequested().isServingDC() && detailRequest.getInventoryRelationshipsRequested().isLocalStore()) {
        if (detailRequest.getInventoryRelationshipsRequested().isServingDC()) {
            List<SiteRelationDTO> siteRelationDTOList = availabilityServiceContextV3.getRequestingSite().getSiteRelationDTOList();
            if (siteRelationDTOList != null && siteRelationDTOList.size() >0) {
                try {
                    primaryDC = siteRelationDTOList.stream()
                            .filter(siteRelationDTO -> siteRelationDTO.getRelationshipType().equalsIgnoreCase("Primary DC"))
                            .map(siteRelationDTO -> {
                                RelatedSite relatedSite = new RelatedSite();
                                relatedSite.setSiteId(siteRelationDTO.getRelatedSiteIdentifier());
                                relatedSite.setType(siteRelationDTO.getRelatedSiteType());
                                addSiteRelation(siteRelationDTO.getRelatedSiteIdentifier(), siteRelationDTO.getRelationshipType(), siteRelationDTO.getRelationshipRank(), availabilityServiceContextV3.getSiteRelationMap());
                                return relatedSite;
                            }).findFirst().get();
                }catch(Exception e){
                    SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest());
                    LOGGER.error(splunkLogger.toString() + " Invalid Site ID for Primary DC");
                }
            }
        }

/*        if (!detailRequest.getInventoryRelationshipsRequested().isLocalStore()) {
            RelatedSite relatedSite = new RelatedSite();
            relatedSite.setSiteId(availabilityServiceContext.getRequestingSite().getSiteIdentifier());
            primaryDC = relatedSite;
        }*/

        availabilityServiceContextV3.setPrimaryDC(primaryDC);
    }


    private void addSiteRelation(String siteIdentifier, String relationshipType, Integer relationshipRank, Map<String, List<SiteRelationship>> siteRelationshipMap) {

        SiteRelationship siteRelationship = new SiteRelationship(relationshipRank, relationshipType);
        if (siteRelationship.getSiteRelationshipRank() != null && siteRelationship.getSiteRelationshipRank() == 99) {
            siteRelationship.setSiteRelationshipRank(null);
        }
        if(siteRelationship.getSiteRelationshipType() != null) {
            List<SiteRelationship> siteRelationshipList = siteRelationshipMap.get(siteIdentifier);
            if (siteRelationshipList == null) {
                siteRelationshipList = new ArrayList<>();
                siteRelationshipMap.put(siteIdentifier, siteRelationshipList);
            }
            siteRelationshipList.add(siteRelationship);
        }
    }
}
